#include <stdlib.h>
#include "imports.h"

#include "mcp.bin.h"
#include "mcp_syms.h"

extern char __bss_start;
extern char __bss_end;

static const uint8_t repairData_set_fault_behavior[] = {
    0xE1, 0x2F, 0xFF, 0x1E, 0xE9, 0x2D, 0x40, 0x30, 0xE5, 0x93, 0x20, 0x00, 0xE1, 0xA0, 0x40, 0x00,
    0xE5, 0x92, 0x30, 0x54, 0xE1, 0xA0, 0x50, 0x01, 0xE3, 0x53, 0x00, 0x01, 0x0A, 0x00, 0x00, 0x02,
    0xE1, 0x53, 0x00, 0x00, 0xE3, 0xE0, 0x00, 0x00, 0x18, 0xBD, 0x80, 0x30, 0xE3, 0x54, 0x00, 0x0D,
};

static const uint8_t repairData_set_panic_behavior[] = {
    0x08, 0x16, 0x6C, 0x00, 0x00, 0x00, 0x18, 0x0C, 0x08, 0x14, 0x40, 0x00, 0x00, 0x00, 0x9D, 0x70,
    0x08, 0x16, 0x84, 0x0C, 0x00, 0x00, 0xB4, 0x0C, 0x00, 0x00, 0x01, 0x01, 0x08, 0x14, 0x40, 0x00,
    0x08, 0x15, 0x00, 0x00, 0x08, 0x17, 0x21, 0x80, 0x08, 0x17, 0x38, 0x00, 0x08, 0x14, 0x30, 0xD4,
    0x08, 0x14, 0x12, 0x50, 0x08, 0x14, 0x12, 0x94, 0xE3, 0xA0, 0x35, 0x36, 0xE5, 0x93, 0x21, 0x94,
    0xE3, 0xC2, 0x2E, 0x21, 0xE5, 0x83, 0x21, 0x94, 0xE5, 0x93, 0x11, 0x94, 0xE1, 0x2F, 0xFF, 0x1E,
    0xE5, 0x9F, 0x30, 0x1C, 0xE5, 0x9F, 0xC0, 0x1C, 0xE5, 0x93, 0x20, 0x00, 0xE1, 0xA0, 0x10, 0x00,
    0xE5, 0x92, 0x30, 0x54, 0xE5, 0x9C, 0x00, 0x00,
};

static const uint8_t repairData_pad_root_thread[] = {
    0xe5, 0x9f, 0x00, 0x44, 0xeb, 0x01, 0xfa, 0x9b, 0xea, 0xff, 0xff, 0x49,
    0x11, 0xfc, 0x00, 0x00, 0x11, 0xfc, 0x00, 0x0c, 0x11, 0xfd, 0x50, 0x00,
    0x11, 0xfd, 0x50, 0x1c, 0x12, 0x14, 0x29, 0x2c, 0x11, 0xfc, 0x02, 0x64,
    0x11, 0xfc, 0x00, 0xd0, 0x11, 0xfc, 0x01, 0x2c, 0x11, 0xfc, 0x02, 0x38,
    0x11, 0xfc, 0x00, 0x68, 0x11, 0xfc, 0x00, 0x30, 0x11, 0xfc, 0x01, 0x50,
    0x11, 0xfc, 0x01, 0x90, 0x11, 0xfc, 0x01, 0xc8, 0x11, 0xfc, 0x02, 0x00,
    0x11, 0xfc, 0x00, 0x98, 0x11, 0xfc, 0x00, 0xf0, 0xe5, 0x2d, 0xe0, 0x04,
    0xeb, 0x02, 0x07, 0xf5, 0xe3, 0xa0, 0x20, 0x03, 0xe3, 0xa0, 0x10, 0x00,
    0xe5, 0x9f, 0x00, 0x44, 0xeb, 0x01, 0xf9, 0xb0
};

#define THUMB_BL(addr, func) ((0xF000F800 | ((((uint32_t)(func) - (uint32_t)(addr) - 4) >> 1) & 0x0FFF)) | ((((uint32_t)(func) - (uint32_t)(addr) - 4) << 4) & 0x7FFF000)) // +-4MB

int _main()
{
    kernel_flush_dcache(0x081200F0, 0x4001); // giving a size >= 0x4000 flushes all cache

    int level = kernel_disable_interrupts();

    unsigned int control_register = disable_mmu();

    // clear our bss
    kernel_memset(&__bss_start, 0, &__bss_end - &__bss_start);

    // recover memory we've overwritten
    kernel_memcpy((void*) 0x081298BC, repairData_set_fault_behavior, sizeof(repairData_set_fault_behavior));
    kernel_memcpy((void*) 0x081296E4, repairData_set_panic_behavior, sizeof(repairData_set_panic_behavior));
    kernel_memcpy((void*) 0x11f00558, repairData_pad_root_thread, sizeof(repairData_pad_root_thread));

    // load the custom ios_mcp
    kernel_memcpy((void *) (_text_start - 0x05000000 + 0x081C0000), mcp, mcp_size);
    kernel_memset((void *) (_bss_start  - 0x05000000 + 0x081C0000), 0, _bss_end - _bss_start);

    // patch out mcp FSA_Mount file descriptor check
    *(volatile uint32_t *) (0x050397d8 - 0x05000000 + 0x081C0000) = 0xe3a00001; // mov r0, #1

    // replace the next loaded rpx
    *(volatile uint32_t *) (0x050254D6 - 0x05000000 + 0x081C0000) = THUMB_BL(0x050254D6, MCP_LoadFile_patch);

    // sd access
    *(volatile uint32_t *) (0x0501dd78 - 0x05000000 + 0x081C0000) = THUMB_BL(0x0501dd78, MCP_ReadCOSXml_patch);
    *(volatile uint32_t *) (0x051105ce - 0x05000000 + 0x081C0000) = THUMB_BL(0x051105ce, MCP_ReadCOSXml_patch);

    // allow any region title launch
    *(volatile uint32_t*)(0xE0030498 - 0xE0000000 + 0x12900000) = 0xE3A00000; // mov r0, #0

    restore_mmu(control_register);

    kernel_invalidate_dcache(0x081298BC, 0x4001); // giving a size >= 0x4000 invalidates all cache
    kernel_invalidate_icache();

    kernel_enable_interrupts(level);

    return 0;
}